import _ from 'lodash';
import { Comment, DeclarationReflection, ParameterReflection, SignatureReflection } from 'typedoc';
import { CommandMethodDeclarationReflection, CommentSource, Example, KnownMethods } from '../converter';
import { AppiumPluginLogger } from '../logger';
import { AllowedHttpMethod, Command, Route } from './types';
/**
 * Abstract representation of metadata for some sort of Appium command
 */
export declare abstract class BaseCommandData {
    /**
     * Loggher
     */
    protected readonly log: AppiumPluginLogger;
    /**
     * The method name of the command handler.
     */
    readonly command: string;
    /**
     * The comment to display for the command, if any exists
     */
    readonly comment?: Comment;
    /**
     * Code to describe how and where {@linkcode comment} came from.
     *
     * For debugging purposes mainly
     */
    readonly commentSource?: CommentSource;
    /**
     * Language-specific examples for the command, if any
     */
    readonly examples?: Example[];
    /**
     * If `true`, this command is from a Plugin (not a Driver)
     */
    readonly isPluginCommand: boolean;
    /**
     * Map of known builtin methods
     */
    readonly knownBuiltinMethods?: KnownMethods;
    /**
     * Actual method reflection.
     */
    readonly methodRefl: CommandMethodDeclarationReflection;
    /**
     * List of optional parameter names derived from a method map
     */
    readonly optionalParams?: string[];
    /**
     * Parameter reflections for this command's method declaration, to eventually be displayed in rendered docs
     *
     * These are _not_ the same objects as in the `parameters` property of a the call signature
     * reflection in `methodRefl`; the comments therein have been aggregated and the parameters have
     * been renamed and possibly truncated.
     */
    readonly parameters?: ParameterReflection[];
    /**
     * The thing which the method is a member of for documentation purposes
     */
    readonly parentRefl?: DeclarationReflection;
    /**
     * List of required parameter names derived from a method map
     */
    readonly requiredParams?: string[];
    /**
     * Signature reflection for this command's method declaration, to eventually be displayed in
     * rendered docs
     *
     * `methodRefl` is a {@linkcode CommandMethodDeclarationReflection}, so it returns a `Promise<T>`, by
     * definition.  This signature reflection is modified so that it returns `T` instead, since
     * `Promise`s don't make much sense in the rendered documentaion.
     *
     * The default TypeDoc output uses the original `SignatureReflection`, so you _will_ see
     * `Promise<T>` there.
     */
    readonly signature?: SignatureReflection;
    /**
     * Returns a list of `ParameterReflection` objects in the command's method declaration;
     * rewrites them to prefer the method map parameter list (and the param names)
     **/
    static rewriteParameters: ((cmd: BaseCommandData) => ParameterReflection[] | undefined) & _.MemoizedFunction;
    /**
     * Rewrites a method's return value for documentation.
     *
     * Given a command having a method declaration, creates a clone of its call signature wherein the
     * return type is unwrapped from `Promise`.  In other words, if a method returns `Promise<T>`,
     * this changes the return type in the signature to `T`.
     *
     * Note that the return type of a command's method declaration should always be a `ReferenceType` having
     * name `Promise`.
     */
    static rewriteSignature: ((cmd: BaseCommandData) => SignatureReflection | undefined) & _.MemoizedFunction;
    constructor(log: AppiumPluginLogger, command: Command, methodRefl: CommandMethodDeclarationReflection, opts?: CommandDataOpts);
    /**
     * Returns `true` if the method or execute map defined parameters for this command
     */
    get hasCommandParams(): boolean;
    /**
     * Should create a shallow clone of the implementing instance
     * @param opts New options to pass to the new instance
     */
    abstract clone(opts: CommandDataOpts): BaseCommandData;
}
/**
 * Options for {@linkcode CommandData} and {@linkcode ExecMethodData} constructors
 */
export interface CommandDataOpts {
    /**
     * The comment to display for the command, if any exists
     */
    comment?: Comment;
    /**
     * Name of the reference which the comment is derived from.
     *
     * For debugging purposes mainly
     */
    commentSource?: CommentSource;
    /**
     * If `true`, `refl` represents a `PluginCommand`, wherein we will ignore
     * the first two parameters altogether.
     */
    isPluginCommand?: boolean;
    /**
     * Known methods in the project
     */
    knownBuiltinMethods?: KnownMethods;
    /**
     * List of optional parameter names derived from a method map
     */
    optionalParams?: string[];
    /**
     * The thing which the method is a member of for documentation purposes
     */
    parentRefl?: DeclarationReflection;
    /**
     * List of required parameter names derived from a method map
     */
    requiredParams?: string[];
}
/**
 * Represents a generic WD or Appium-specific endpoint
 */
export declare class CommandData extends BaseCommandData {
    /**
     * The HTTP method of the route
     */
    readonly httpMethod: AllowedHttpMethod;
    /**
     * The route of the command
     */
    readonly route: Route;
    constructor(log: AppiumPluginLogger, command: Command, methodRefl: CommandMethodDeclarationReflection, httpMethod: AllowedHttpMethod, route: Route, opts?: CommandDataOpts);
    /**
     * Creates a **shallow** clone of this instance.
     *
     * Keeps props {@linkcode BaseCommandData.command command},
     * {@linkcode CommandData.httpMethod httpMethod} and
     * {@linkcode CommandData.route route}, then applies any other options.
     * @param opts Options to apply. _Note:_ you probably want to provide a new `parentRefl`.
     * @returns Cloned instance
     */
    clone(methodRefl: CommandMethodDeclarationReflection, opts?: CommandDataOpts): CommandData;
}
/**
 * Represents an "execute command" ("execute method")
 *
 * Each will have a unique `script` property which is provided as the script to run via the
 * `execute` WD endpoint.
 *
 * All of these share the same `execute` route, so it is omitted from this interface.
 */
export declare class ExecMethodData extends BaseCommandData {
    /**
     * The name/identifier of the execute script
     *
     * This is different than the method name.
     */
    readonly script: string;
    constructor(log: AppiumPluginLogger, command: Command, methodRefl: CommandMethodDeclarationReflection, script: string, opts?: CommandDataOpts);
    /**
     * Creates a **shallow** clone of this instance.
     *
     * Keeps props {@linkcode BaseCommandData.command command}, {@linkcode ExecMethod.script script},
     * then applies any other options.
     * @param opts Options to apply
     * @returns Cloned instance
     */
    clone(opts: CommandDataOpts): ExecMethodData;
}
//# sourceMappingURL=command-data.d.ts.map